"""
Dual purpose test:
1. Test populating the account with typical scenario and test runs
2. A re-runnable script to populate a demo account

"""

import os
from typing import Any
from urllib.parse import urlparse

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.common import BASE_URL
from okareo.model_under_test import CustomModel, OpenAIModel
from okareo_api_client.models import ScenarioSetResponse, TestRunType


@pytest.fixture
def non_mocked_hosts() -> list:
    return [urlparse(BASE_URL).hostname]


@pytest.fixture
def rnd() -> str:
    return random_string(5)


@pytest.fixture
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


def test_load_classification(okareo: Okareo, rnd: str) -> None:
    seed: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=os.path.join(os.path.dirname(__file__), "webbizz_class_seed.jsonl"),
        scenario_name=f"Support - Seed Upload {rnd}",
    )

    assert seed.scenario_id
    assert seed.project_id
    assert seed.time_created

    rephrase: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=os.path.join(
            os.path.dirname(__file__), "webbizz_class_rephrase.jsonl"
        ),
        scenario_name=f"Support - Generated Rephrase {rnd}",
    )

    assert rephrase.scenario_id
    assert rephrase.project_id
    assert rephrase.time_created

    conditional: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=os.path.join(
            os.path.dirname(__file__), "webbizz_class_conditional.jsonl"
        ),
        scenario_name=f"Support - Generated Conditional {rnd}",
    )

    assert conditional.scenario_id
    assert conditional.project_id
    assert conditional.time_created

    class RetrievalModel(CustomModel):
        # Callable to be applied to each scenario in the scenario set
        def invoke(self, input_value: str) -> Any:
            # call your model being tested here using <input> from the scenario set

            # mock code returnign a random label
            labels = [
                "returns",
                "complaints",
                "pricing",
                "product",
                "shipping",
                "support",
            ]
            import random

            actual = random.choice(labels)

            return actual, {
                "labels": actual,
                "confidence": round(random.uniform(0.30, 0.99), 2),
            }

    model_under_test = okareo.register_model(
        name="support_intent_classifier",
        model=RetrievalModel(name="custom classification"),
    )

    test_run_name = f"Support - Rephrase Test Run {rnd}"

    rephrase_test_run = model_under_test.run_test(
        scenario=rephrase,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_CLASS_CLASSIFICATION,
    )

    assert rephrase_test_run.id
    assert rephrase_test_run.model_metrics

    test_run_name = f"Support - Conditional Test Run {rnd}"

    conditional_test_run = model_under_test.run_test(
        scenario=conditional,
        name=test_run_name,
        test_run_type=TestRunType.MULTI_CLASS_CLASSIFICATION,
    )

    assert conditional_test_run.id
    assert conditional_test_run.model_metrics


def test_load_retrieval(okareo: Okareo, rnd: str) -> None:
    file_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "examples",
        "webbizz_10_articles.jsonl",
    )
    articles: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=file_path,
        scenario_name=f"Support - Articles Seed {rnd}",
    )

    assert articles.scenario_id
    assert articles.project_id
    assert articles.time_created

    questions: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=os.path.join(
            os.path.dirname(__file__), "webbizz_retrieval_questions.jsonl"
        ),
        scenario_name=f"Support - Generated Questions {rnd}",
    )

    assert questions.scenario_id
    assert questions.project_id
    assert questions.time_created

    class RetrievalModel(CustomModel):
        def invoke(self, input_value: str) -> Any:
            # call your embedding model and vector db retrieval being tested here using <input> from the scenario set
            # we are using a random response here for demonstration purposes
            article_ids = [
                "75eaa363-dfcc-499f-b2af-1407b43cb133",
                "ac0d464c-f673-44b8-8195-60c965e47525",
                "35a4fd5b-453e-4ca6-9536-f20db7303344",
                "a8a97b0e-8d9a-4a1c-b93e-83d2bc9e5266",
                "0b85c12f-6ea6-4d4a-85de-6c6e9a9f8c78",
                "cda67f1d-19f2-4b45-9f3e-3b8d67f8c6c5",
                "6e4f1c97-3f7a-4fcd-a4a3-69c9817c8fd1",
                "f658c264-4a8a-4c93-a6d7-9a3d75f5a6f3",
                "aacf7a34-9d3a-4e2a-9a5c-91f2a0e8a12d",
                "f1a37b5e-58c4-4f5a-bc42-1b70253b8bf3",
            ]
            import random

            selected_ids = random.sample(article_ids, 5)
            rounded_random_scores = sorted(
                [round(random.random(), 2) for _ in range(5)], reverse=True
            )

            # higher score value means more relevant
            parsed_ids_with_scores = [
                (selected_id, score)
                for selected_id, score in zip(selected_ids, rounded_random_scores)
            ]

            model_response = {"matches": "additional context from the model"}

            # return a tuple of (parsed_ids_with_scores, overall model response context)
            return parsed_ids_with_scores, model_response

    model_under_test = okareo.register_model(
        name="vectordb_retrieval",
        model=RetrievalModel(name="custom retrieval"),
    )

    test_run_name = f"Support - Retrieval Test Run {rnd}"

    retrieval_test_run = model_under_test.run_test(
        scenario=questions,
        name=test_run_name,
        test_run_type=TestRunType.INFORMATION_RETRIEVAL,
    )

    assert retrieval_test_run.id
    assert retrieval_test_run.model_metrics


TEST_SUMMARIZE_TEMPLATE = """
Provide a brief summary of the following paragraph of text:

{input}

Summary:

"""


def test_load_generation(okareo: Okareo, rnd: str) -> None:
    file_path = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        "examples",
        "webbizz_10_articles.jsonl",
    )
    scenario = okareo.upload_scenario_set(
        file_path=file_path, scenario_name=f"openai-scenario-set-{rnd}"
    )

    mut = okareo.register_model(
        name=f"openai-ci-run-{rnd}",
        model=OpenAIModel(
            model_id="gpt-3.5-turbo",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )

    run_resp = mut.run_test(
        name=f"openai-chat-run-{rnd}",
        scenario=scenario,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-{rnd}"



""" Contains all the data models used in inputs/outputs """

from .body_evaluator_delete_v0_evaluator_evaluator_id_delete import BodyEvaluatorDeleteV0EvaluatorEvaluatorIdDelete
from .body_evaluator_upload_v0_evaluator_upload_post import BodyEvaluatorUploadV0EvaluatorUploadPost
from .body_scenario_sets_upload_v0_scenario_sets_upload_post import BodyScenarioSetsUploadV0ScenarioSetsUploadPost
from .datapoint_list_item import DatapointListItem
from .datapoint_list_item_input_type_0 import DatapointListItemInputType0
from .datapoint_list_item_result_type_0 import DatapointListItemResultType0
from .datapoint_response import DatapointResponse
from .datapoint_schema import DatapointSchema
from .datapoint_search import DatapointSearch
from .datapoint_summary_item import DatapointSummaryItem
from .error_response import ErrorResponse
from .evaluator_brief_response import EvaluatorBriefResponse
from .evaluator_detailed_response import EvaluatorDetailedResponse
from .evaluator_generate_response import EvaluatorGenerateResponse
from .evaluator_spec_request import EvaluatorSpecRequest
from .feedback_range_summary import FeedbackRangeSummary
from .find_test_data_point_payload import FindTestDataPointPayload
from .general_find_payload import GeneralFindPayload
from .generation_tone import GenerationTone
from .model_under_test_response import ModelUnderTestResponse
from .model_under_test_schema import ModelUnderTestSchema
from .model_under_test_schema_models import ModelUnderTestSchemaModels
from .model_under_test_schema_models_additional_property import ModelUnderTestSchemaModelsAdditionalProperty
from .project_response import ProjectResponse
from .project_schema import ProjectSchema
from .scenario_data_poin_response import ScenarioDataPoinResponse
from .scenario_data_poin_response_input_type_0 import ScenarioDataPoinResponseInputType0
from .scenario_data_poin_response_result_type_0 import ScenarioDataPoinResponseResultType0
from .scenario_set_create import ScenarioSetCreate
from .scenario_set_generate import ScenarioSetGenerate
from .scenario_set_response import ScenarioSetResponse
from .scenario_set_update import ScenarioSetUpdate
from .scenario_type import ScenarioType
from .seed_data import SeedData
from .semantic_payload import SemanticPayload
from .semantic_result import SemanticResult
from .summary_datapoint_search import SummaryDatapointSearch
from .test_data_point_item import TestDataPointItem
from .test_data_point_item_metric_value import TestDataPointItemMetricValue
from .test_data_point_payload import TestDataPointPayload
from .test_data_point_response import TestDataPointResponse
from .test_run_item import TestRunItem
from .test_run_item_model_metrics import TestRunItemModelMetrics
from .test_run_payload import TestRunPayload
from .test_run_payload_v2 import TestRunPayloadV2
from .test_run_payload_v2_api_keys import TestRunPayloadV2ApiKeys
from .test_run_payload_v2_metrics_kwargs import TestRunPayloadV2MetricsKwargs
from .test_run_payload_v2_model_results import TestRunPayloadV2ModelResults
from .test_run_type import TestRunType

__all__ = (
    "BodyEvaluatorDeleteV0EvaluatorEvaluatorIdDelete",
    "BodyEvaluatorUploadV0EvaluatorUploadPost",
    "BodyScenarioSetsUploadV0ScenarioSetsUploadPost",
    "DatapointListItem",
    "DatapointListItemInputType0",
    "DatapointListItemResultType0",
    "DatapointResponse",
    "DatapointSchema",
    "DatapointSearch",
    "DatapointSummaryItem",
    "ErrorResponse",
    "EvaluatorBriefResponse",
    "EvaluatorDetailedResponse",
    "EvaluatorGenerateResponse",
    "EvaluatorSpecRequest",
    "FeedbackRangeSummary",
    "FindTestDataPointPayload",
    "GeneralFindPayload",
    "GenerationTone",
    "ModelUnderTestResponse",
    "ModelUnderTestSchema",
    "ModelUnderTestSchemaModels",
    "ModelUnderTestSchemaModelsAdditionalProperty",
    "ProjectResponse",
    "ProjectSchema",
    "ScenarioDataPoinResponse",
    "ScenarioDataPoinResponseInputType0",
    "ScenarioDataPoinResponseResultType0",
    "ScenarioSetCreate",
    "ScenarioSetGenerate",
    "ScenarioSetResponse",
    "ScenarioSetUpdate",
    "ScenarioType",
    "SeedData",
    "SemanticPayload",
    "SemanticResult",
    "SummaryDatapointSearch",
    "TestDataPointItem",
    "TestDataPointItemMetricValue",
    "TestDataPointPayload",
    "TestDataPointResponse",
    "TestRunItem",
    "TestRunItemModelMetrics",
    "TestRunPayload",
    "TestRunPayloadV2",
    "TestRunPayloadV2ApiKeys",
    "TestRunPayloadV2MetricsKwargs",
    "TestRunPayloadV2ModelResults",
    "TestRunType",
)



import os

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import CohereModel, OpenAIModel, PineconeDb, QdrantDB
from okareo_api_client.models import ScenarioSetResponse
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.scenario_type import ScenarioType
from okareo_api_client.models.seed_data import SeedData
from okareo_api_client.models.test_run_item import TestRunItem
from okareo_api_client.models.test_run_type import TestRunType
from okareo_api_client.types import Unset


@pytest.fixture(scope="module")
def rnd() -> str:
    return random_string(5)


@pytest.fixture(scope="module")
def okareo() -> Okareo:
    return Okareo(api_key=API_KEY)


TEST_SUMMARIZE_TEMPLATE = """
Provide a brief summary of the following paragraph of text:

{input}

Summary:

"""


@pytest.fixture(scope="module")
def article_scenario_set(rnd: str, okareo: Okareo) -> ScenarioSetResponse:
    file_path = os.path.join(os.path.dirname(__file__), "webbizz_3_test_article.jsonl")
    articles: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=file_path, scenario_name=f"openai-scenario-set-{rnd}"
    )

    return articles


def test_run_test_openai(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    mut = okareo.register_model(
        name=f"openai-ci-run-{rnd}",
        model=OpenAIModel(
            model_id="gpt-3.5-turbo",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=None,
        ),
    )

    run_resp = mut.run_test(
        name=f"openai-chat-run-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-{rnd}"
    assert_metrics(run_resp)


def test_run_test_openai_2prompts(
    rnd: str, okareo: Okareo, article_scenario_set: ScenarioSetResponse
) -> None:
    mut2 = okareo.register_model(
        name=f"openai-ci-run-{rnd}",
        model=OpenAIModel(
            model_id="gpt-3.5-turbo",
            temperature=0,
            system_prompt_template=TEST_SUMMARIZE_TEMPLATE,
            user_prompt_template=TEST_SUMMARIZE_TEMPLATE,
        ),
    )

    run_resp = mut2.run_test(
        name=f"openai-chat-run-{rnd}",
        scenario=article_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        calculate_metrics=True,
    )
    assert run_resp.name == f"openai-chat-run-{rnd}"
    assert_metrics(run_resp)


def assert_metrics(run_resp: TestRunItem) -> None:
    assert run_resp.model_metrics is not None and not isinstance(
        run_resp.model_metrics, Unset
    )
    metrics_dict = run_resp.model_metrics.to_dict()

    assert metrics_dict["mean_scores"] is not None
    assert_scores(metrics_dict["mean_scores"])
    assert metrics_dict["scores_by_row"] is not None
    assert len(metrics_dict["scores_by_row"]) == 3
    for row in metrics_dict["scores_by_row"]:
        assert_scores(row)


def assert_scores(scores: dict) -> None:
    dimension_keys = ["consistency", "coherence", "fluency", "relevance"]
    for dimension in dimension_keys:
        assert dimension in scores
        assert isinstance(scores[dimension], float)
        assert 1 <= scores[dimension] <= 5


def test_run_test_cohere(rnd: str, okareo: Okareo) -> None:
    seed_data = [
        SeedData(input_="what are you able to set up in aws?", result="capabilities"),
        SeedData(
            input_="what's the procedure to request for more information?",
            result="general",
        ),
        SeedData(
            input_="what are the steps to deploy on heroku?", result="capabilities"
        ),
    ]

    scenario_set_create = ScenarioSetCreate(
        name=f"cohere-test-ci-{rnd}",
        number_examples=1,
        seed_data=seed_data,
        generation_type=ScenarioType.REPHRASE_INVARIANT,
    )
    scenario = okareo.create_scenario_set(scenario_set_create)

    mut = okareo.register_model(
        name=f"classification-cohere-ci-{rnd}",
        model=CohereModel(
            model_id="2386d4d1-c617-4183-8c87-5550c7f222e6-ft",
            model_type="classify",
        ),
    )

    run_resp = mut.run_test(
        name=f"cohere-classification-run-{rnd}",
        scenario=scenario,
        api_key=os.environ["COHERE_API_KEY"],
        calculate_metrics=True,
    )
    assert run_resp.name == f"cohere-classification-run-{rnd}"


@pytest.fixture(scope="module")
def question_scenario_set(rnd: str, okareo: Okareo) -> ScenarioSetResponse:
    file_path = os.path.join(
        os.path.dirname(__file__), "webbizz_retrieval_questions.jsonl"
    )
    articles: ScenarioSetResponse = okareo.upload_scenario_set(
        file_path=file_path, scenario_name=f"ci-pinecone-cohere-{rnd}"
    )

    return articles


def test_run_test_cohere_pinecone_ir(
    rnd: str, okareo: Okareo, question_scenario_set: ScenarioSetResponse
) -> None:
    mut = okareo.register_model(
        name=f"ci-pinecone-cohere-english-light-v3.0-{rnd}",
        model=[
            CohereModel(
                model_id="embed-english-light-v3.0",
                model_type="embed",
                input_type="search_query",
            ),
            PineconeDb(
                index_name="my-test-index",
                region="gcp-starter",
                project_id="kwnp6kx",
                top_k=3,
            ),
        ],
    )

    run_resp = mut.run_test(
        name=f"ci-pinecone-cohere-embed-{rnd}",
        scenario=question_scenario_set,
        calculate_metrics=True,
        test_run_type=TestRunType.INFORMATION_RETRIEVAL,
        api_keys={
            "cohere": os.environ["COHERE_API_KEY"],
            "pinecone": os.environ["PINECONE_API_KEY"],
        },
        metrics_kwargs={
            "mrr_at_k": [2, 4, 8],
            "map_at_k": [1, 2],
        },
    )
    assert run_resp.name == f"ci-pinecone-cohere-embed-{rnd}"


def test_run_test_cohere_qdrant_ir(
    rnd: str, okareo: Okareo, question_scenario_set: ScenarioSetResponse
) -> None:
    mut = okareo.register_model(
        name=f"ci-qdrant-cohere-english-light-v3.0-{rnd}",
        model=[
            CohereModel(
                model_id="embed-english-light-v3.0",
                model_type="embed",
                input_type="search_query",
            ),
            QdrantDB(
                collection_name="ci_test_collection",
                url="https://366662aa-e06e-4d40-a1d0-dc6aedbef44e.us-east4-0.gcp.cloud.qdrant.io:6333",
                top_k=10,
            ),
        ],
    )

    run_resp = mut.run_test(
        name=f"ci-qdrant-cohere-embed-{rnd}",
        scenario=question_scenario_set,
        calculate_metrics=True,
        test_run_type=TestRunType.INFORMATION_RETRIEVAL,
        api_keys={
            "cohere": os.environ["COHERE_API_KEY"],
            "qdrant": os.environ["QDRANT_API_KEY"],
        },
        metrics_kwargs={
            "mrr_at_k": [2, 4, 8, 16],
            "map_at_k": [1, 3, 5, 10, 20],
        },
    )
    assert run_resp.name == f"ci-qdrant-cohere-embed-{rnd}"



from typing import List

import pytest
from okareo_tests.common import API_KEY

from okareo import Okareo


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


def test_get_projects(okareo_client: Okareo) -> None:
    projects = okareo_client.get_projects()
    assert projects
    assert isinstance(projects, List)
    assert len(projects) > 0
    assert projects[0].id
    assert projects[0].name



from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="SemanticPayload")


@_attrs_define
class SemanticPayload:
    """
    Attributes:
        source_text (str):
        output_text (str):
    """

    source_text: str
    output_text: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        source_text = self.source_text
        output_text = self.output_text

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "source_text": source_text,
                "output_text": output_text,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        source_text = d.pop("source_text")

        output_text = d.pop("output_text")

        semantic_payload = cls(
            source_text=source_text,
            output_text=output_text,
        )

        semantic_payload.additional_properties = d
        return semantic_payload

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class GenerationTone(str, Enum):
    EMPATHETIC = "Empathetic"
    FORMAL = "Formal"
    INFORMAL = "Informal"
    NEUTRAL = "Neutral"
    PERSUASIVE = "Persuasive"

    def __str__(self) -> str:
        return str(self.value)



import collections
import threading
import time
from abc import abstractmethod
from typing import Any, Callable, Deque, Tuple, Union

from okareo_api_client.client import Client

_DEFAULT_MAX_QUEUE_SIZE = 2048
_DEFAULT_SCHEDULE_DELAY_MILLIS = 1000
_DEFAULT_MAX_BATCH_SIZE = 512
_DEFAULT_ASYNC_CALL_RETRIES = 5


class AsyncProcessorMixin:
    def __init__(self, name: str = "AsyncProcessor") -> None:
        self.queue: Deque[Tuple[Callable, Any]] = collections.deque(
            [], _DEFAULT_MAX_QUEUE_SIZE
        )
        self.done = False
        self._data_dropped = False
        self.worker_thread = threading.Thread(
            name=name, target=self.worker, daemon=True
        )
        self.condition = threading.Condition(threading.Lock())
        self.worker_thread.start()

    @abstractmethod
    def get_client(self) -> Client:
        """Return API Client instance"""

    @abstractmethod
    def get_api_key(self) -> str:
        """Get Okareo API key to use"""

    def async_call(self, func: Callable, data: Any) -> bool:
        if len(self.queue) == _DEFAULT_MAX_QUEUE_SIZE:
            if not self._data_dropped:
                print("Queue is full, data points might get dropped.")
            self._data_dropped = True
        self.queue.append((func, data))

        if len(self.queue) >= int(_DEFAULT_MAX_BATCH_SIZE):
            with self.condition:
                self.condition.notify()
        return True

    def _perform_call(self, func: Callable, data: Any) -> Union[Any, None]:
        attempts = 0
        while attempts < _DEFAULT_ASYNC_CALL_RETRIES:
            try:
                return func(
                    client=self.get_client(),
                    api_key=self.get_api_key(),
                    json_body=data,
                )

            except Exception as e:
                print("Error performing async call", e)
                attempts += 1
            # .1, .2, .4, .8, 1.6, 3.2, ...
            time.sleep(0.1 * 2 ** (attempts - 1))
        return None

    def worker(self) -> None:
        timeout = _DEFAULT_SCHEDULE_DELAY_MILLIS / 1e3

        while not self.done:
            with self.condition:
                if self.done:
                    break  # if flag has changed
                if not self.queue:
                    self.condition.wait(timeout)

            idx = 0
            entries = []
            while idx < _DEFAULT_MAX_BATCH_SIZE and self.queue:
                entries.append(self.queue.popleft())

            for entry in entries:
                func, data = entry
                self._perform_call(func, data)

    def flush(self) -> None:
        # signal the worker thread to finish and then wait for it
        print("Shutting down")
        self.done = True
        with self.condition:
            self.condition.notify_all()
        self.worker_thread.join()



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.test_run_type import TestRunType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.test_run_payload_v2_api_keys import TestRunPayloadV2ApiKeys
    from ..models.test_run_payload_v2_metrics_kwargs import TestRunPayloadV2MetricsKwargs
    from ..models.test_run_payload_v2_model_results import TestRunPayloadV2ModelResults


T = TypeVar("T", bound="TestRunPayloadV2")


@_attrs_define
class TestRunPayloadV2:
    """
    Attributes:
        mut_id (str): ID of the model
        scenario_id (str): ID of the scenario set
        api_keys (Union[Unset, TestRunPayloadV2ApiKeys]): Dictionary that maps model type to the respective API keys
        metrics_kwargs (Union[Unset, TestRunPayloadV2MetricsKwargs]): Dictionary of metrics to be measured
        name (Union[Unset, str]): Name of the test run
        type (Union[Unset, TestRunType]): An enumeration. Default: TestRunType.MULTI_CLASS_CLASSIFICATION.
        calculate_metrics (Union[Unset, bool]): Boolean value indicating if metrics should be calculated for the test
            run Default: True.
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter test runs in the Okareo app
        project_id (Union[Unset, str]): ID of the project
        model_results (Union[Unset, TestRunPayloadV2ModelResults]):
        checks (Union[Unset, List[str]]): List of checks to include in the test run.
    """

    mut_id: str
    scenario_id: str
    api_keys: Union[Unset, "TestRunPayloadV2ApiKeys"] = UNSET
    metrics_kwargs: Union[Unset, "TestRunPayloadV2MetricsKwargs"] = UNSET
    name: Union[Unset, str] = UNSET
    type: Union[Unset, TestRunType] = TestRunType.MULTI_CLASS_CLASSIFICATION
    calculate_metrics: Union[Unset, bool] = True
    tags: Union[Unset, List[str]] = UNSET
    project_id: Union[Unset, str] = UNSET
    model_results: Union[Unset, "TestRunPayloadV2ModelResults"] = UNSET
    checks: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        mut_id = self.mut_id
        scenario_id = self.scenario_id
        api_keys: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.api_keys, Unset):
            api_keys = self.api_keys.to_dict()

        metrics_kwargs: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.metrics_kwargs, Unset):
            metrics_kwargs = self.metrics_kwargs.to_dict()

        name = self.name
        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        calculate_metrics = self.calculate_metrics
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        project_id = self.project_id
        model_results: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.model_results, Unset):
            model_results = self.model_results.to_dict()

        checks: Union[Unset, List[str]] = UNSET
        if not isinstance(self.checks, Unset):
            checks = self.checks

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "mut_id": mut_id,
                "scenario_id": scenario_id,
            }
        )
        if api_keys is not UNSET:
            field_dict["api_keys"] = api_keys
        if metrics_kwargs is not UNSET:
            field_dict["metrics_kwargs"] = metrics_kwargs
        if name is not UNSET:
            field_dict["name"] = name
        if type is not UNSET:
            field_dict["type"] = type
        if calculate_metrics is not UNSET:
            field_dict["calculate_metrics"] = calculate_metrics
        if tags is not UNSET:
            field_dict["tags"] = tags
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if model_results is not UNSET:
            field_dict["model_results"] = model_results
        if checks is not UNSET:
            field_dict["checks"] = checks

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.test_run_payload_v2_api_keys import TestRunPayloadV2ApiKeys
        from ..models.test_run_payload_v2_metrics_kwargs import TestRunPayloadV2MetricsKwargs
        from ..models.test_run_payload_v2_model_results import TestRunPayloadV2ModelResults

        d = src_dict.copy()
        mut_id = d.pop("mut_id")

        scenario_id = d.pop("scenario_id")

        _api_keys = d.pop("api_keys", UNSET)
        api_keys: Union[Unset, TestRunPayloadV2ApiKeys]
        if isinstance(_api_keys, Unset):
            api_keys = UNSET
        else:
            api_keys = TestRunPayloadV2ApiKeys.from_dict(_api_keys)

        _metrics_kwargs = d.pop("metrics_kwargs", UNSET)
        metrics_kwargs: Union[Unset, TestRunPayloadV2MetricsKwargs]
        if isinstance(_metrics_kwargs, Unset):
            metrics_kwargs = UNSET
        else:
            metrics_kwargs = TestRunPayloadV2MetricsKwargs.from_dict(_metrics_kwargs)

        name = d.pop("name", UNSET)

        _type = d.pop("type", UNSET)
        type: Union[Unset, TestRunType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = TestRunType(_type)

        calculate_metrics = d.pop("calculate_metrics", UNSET)

        tags = cast(List[str], d.pop("tags", UNSET))

        project_id = d.pop("project_id", UNSET)

        _model_results = d.pop("model_results", UNSET)
        model_results: Union[Unset, TestRunPayloadV2ModelResults]
        if isinstance(_model_results, Unset):
            model_results = UNSET
        else:
            model_results = TestRunPayloadV2ModelResults.from_dict(_model_results)

        checks = cast(List[str], d.pop("checks", UNSET))

        test_run_payload_v2 = cls(
            mut_id=mut_id,
            scenario_id=scenario_id,
            api_keys=api_keys,
            metrics_kwargs=metrics_kwargs,
            name=name,
            type=type,
            calculate_metrics=calculate_metrics,
            tags=tags,
            project_id=project_id,
            model_results=model_results,
            checks=checks,
        )

        test_run_payload_v2.additional_properties = d
        return test_run_payload_v2

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import logging
import os
import random
import string
from typing import Any, Dict, List, Optional
from uuid import UUID

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import AgentAction, AgentFinish, LLMResult

from okareo.okareo import Okareo


class CallbackHandler(BaseCallbackHandler):
    """
    Base callback handler that can be used to handle callbacks
    from langchain.
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        mut_name: Optional[str] = None,
        context_token: Optional[str] = None,
    ) -> None:
        """Initialize callback handler."""
        self.okareo = Okareo(api_key or os.environ["OKAREO_API_KEY"])
        self.inputs: List[Dict[str, Any]] = []
        self.context_token = context_token or "".join(
            random.choices(string.ascii_letters, k=10)
        )
        self.registered_model = None
        if mut_name:
            self.registered_model = self.okareo.register_model(
                name=mut_name, tags=["langchain"]
            )

    def on_llm_start(
        self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        """Run when LLM starts running."""
        logging.debug("on_llm_start", serialized, prompts, kwargs)
        self.inputs.append(
            {
                "prompts": prompts,
                "invocation_params": kwargs["invocation_params"],
            }
        )

    def on_llm_new_token(self, token: str, **kwargs: Any) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""
        logging.debug("on_llm_new_token", token, kwargs)

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        logging.debug("on_llm_end", response, kwargs)
        model = self.registered_model
        if not model:
            model_name = (
                response.llm_output.get("model_name", "<unknown>")
                if response.llm_output
                else "<unknown>"
            )
            model = self.okareo.register_model(name=model_name, tags=["langchain"])
        model.add_data_point(
            input_obj=self.inputs.pop(),
            result_obj={
                "outputs": response.generations,
                "llm_output": response.llm_output,
            },
            context_token=self.context_token,
            tags=["langchain"],
        )

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM errors."""

    def on_chain_start(
        self, serialized: Dict[str, Any], inputs: Dict[str, Any], **kwargs: Any
    ) -> Any:
        """Run when chain starts running."""
        logging.debug("on_chain_start", serialized, inputs, kwargs)
        class_name = serialized.get("name", serialized.get("id", ["<unknown>"])[-1])

        if not self.registered_model:
            self.registered_model = self.okareo.register_model(
                name=class_name, tags=["langchain"]
            )
        self.inputs.append(
            {
                "template": serialized.get("kwargs", {})
                .get("prompt", {})
                .get("kwargs", {})
                .get("template", ""),
                "input": inputs,
                "tags": kwargs.get("tags", []),
            }
        )

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        logging.debug("on_chain_end", outputs, kwargs)

        assert self.registered_model
        self.registered_model.add_data_point(
            input_obj=self.inputs.pop(),
            result_obj={
                "outputs": outputs,
                "tags": kwargs.get("tags", []),
            },
            context_token=self.context_token,
            tags=["langchain"],
        )

    def on_chain_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when chain errors."""
        print("\n\033[1m> Chain error\033[0m")
        assert self.registered_model
        self.registered_model.add_data_point(
            input_obj=self.inputs.pop(),
            result_obj={},
            context_token=self.context_token,
            tags=["langchain"],
            error_message=str(error),
        )

    def on_tool_start(
        self, serialized: Dict[str, Any], input_str: str, **kwargs: Any
    ) -> Any:
        """Run when tool starts running."""
        logging.debug("on_tool_start", serialized, input_str, kwargs)

    def on_tool_end(self, output: str, **kwargs: Any) -> Any:
        """Run when tool ends running."""
        logging.debug("on_tool_end", output, kwargs)

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when tool errors."""

    def on_text(self, text: str, **kwargs: Any) -> Any:
        """Run on arbitrary text."""

    def on_agent_action(self, action: AgentAction, **kwargs: Any) -> Any:
        """Run on agent action."""
        logging.debug("on_agent_action", action, kwargs)

    def on_agent_finish(self, finish: AgentFinish, **kwargs: Any) -> Any:
        """Run on agent end."""
        logging.debug("on_agent_action", finish, kwargs)



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.test_run_type import TestRunType
from ..types import UNSET, Unset

T = TypeVar("T", bound="TestRunPayload")


@_attrs_define
class TestRunPayload:
    """
    Attributes:
        mut_id (Union[Unset, str]): ID of model
        scenario_set_id (Union[Unset, str]): ID of scenario set
        name (Union[Unset, str]): Name of test run
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter test runs in the Okareo app
        type (Union[Unset, TestRunType]): An enumeration.
        start_time (Union[Unset, datetime.datetime]):
        end_time (Union[Unset, datetime.datetime]):
        calculate_model_metrics (Union[Unset, bool]): Boolean value indicating if model metrics should be calculated
    """

    mut_id: Union[Unset, str] = UNSET
    scenario_set_id: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    type: Union[Unset, TestRunType] = UNSET
    start_time: Union[Unset, datetime.datetime] = UNSET
    end_time: Union[Unset, datetime.datetime] = UNSET
    calculate_model_metrics: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        mut_id = self.mut_id
        scenario_set_id = self.scenario_set_id
        name = self.name
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        start_time: Union[Unset, str] = UNSET
        if not isinstance(self.start_time, Unset):
            start_time = self.start_time.isoformat()

        end_time: Union[Unset, str] = UNSET
        if not isinstance(self.end_time, Unset):
            end_time = self.end_time.isoformat()

        calculate_model_metrics = self.calculate_model_metrics

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if scenario_set_id is not UNSET:
            field_dict["scenario_set_id"] = scenario_set_id
        if name is not UNSET:
            field_dict["name"] = name
        if tags is not UNSET:
            field_dict["tags"] = tags
        if type is not UNSET:
            field_dict["type"] = type
        if start_time is not UNSET:
            field_dict["start_time"] = start_time
        if end_time is not UNSET:
            field_dict["end_time"] = end_time
        if calculate_model_metrics is not UNSET:
            field_dict["calculate_model_metrics"] = calculate_model_metrics

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        mut_id = d.pop("mut_id", UNSET)

        scenario_set_id = d.pop("scenario_set_id", UNSET)

        name = d.pop("name", UNSET)

        tags = cast(List[str], d.pop("tags", UNSET))

        _type = d.pop("type", UNSET)
        type: Union[Unset, TestRunType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = TestRunType(_type)

        _start_time = d.pop("start_time", UNSET)
        start_time: Union[Unset, datetime.datetime]
        if isinstance(_start_time, Unset):
            start_time = UNSET
        else:
            start_time = isoparse(_start_time)

        _end_time = d.pop("end_time", UNSET)
        end_time: Union[Unset, datetime.datetime]
        if isinstance(_end_time, Unset):
            end_time = UNSET
        else:
            end_time = isoparse(_end_time)

        calculate_model_metrics = d.pop("calculate_model_metrics", UNSET)

        test_run_payload = cls(
            mut_id=mut_id,
            scenario_set_id=scenario_set_id,
            name=name,
            tags=tags,
            type=type,
            start_time=start_time,
            end_time=end_time,
            calculate_model_metrics=calculate_model_metrics,
        )

        test_run_payload.additional_properties = d
        return test_run_payload

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os
from typing import Any, List, Optional

import dotenv
from litellm.integrations.custom_logger import CustomLogger  # type: ignore
from openai._models import BaseModel as OpenAIObject

from okareo import Okareo

dotenv.load_dotenv()


class LiteLLMLogger(CustomLogger):  # type: ignore
    def __init__(
        self,
        api_key: str,
        mut_name: str,
        context_token: str,
        tags: Optional[List[str]] = None,
        host_address: Optional[str] = None,
    ) -> None:
        if host_address and len(host_address) > 0:
            self.okareo = Okareo(api_key, base_path=host_address)
        else:
            self.okareo = Okareo(api_key)

        self.context_token = context_token
        self.registered_model = self.okareo.register_model(name=mut_name, tags=tags)

    def log_success_event(
        self, kwargs: Any, response_obj: Any, start_time: Any, end_time: Any
    ) -> None:
        self.registered_model.add_data_point_async(
            input_obj=self.parse_input_obj(kwargs),
            input_datetime=str(start_time),
            result_obj=self.parse_response_obj(kwargs, response_obj),
            result_datetime=str(end_time),
            context_token=self.context_token,
        )

    async def async_log_success_event(
        self, kwargs: Any, response_obj: Any, start_time: Any, end_time: Any
    ) -> None:
        self.registered_model.add_data_point_async(
            input_obj=self.parse_input_obj(kwargs),
            input_datetime=str(start_time),
            result_obj=self.parse_response_obj(kwargs, response_obj),
            result_datetime=str(end_time),
            context_token=self.context_token,
        )

    def parse_input_obj(self, kwargs: Any) -> dict:
        input_data = {}
        for idx, value in enumerate(kwargs["messages"]):
            input_data[str(idx)] = value
        return {"input": input_data}

    def parse_response_obj(self, kwargs: Any, response_obj: Any) -> Any:
        try:
            if isinstance(response_obj, OpenAIObject):
                formatted_response = dict(response_obj.model_dump())
            else:
                formatted_response = {"raw_response": response_obj}
        except Exception as e:
            formatted_response = {"exception": str(e), "raw_response": response_obj}

        return formatted_response


class LiteLLMProxyLogger(LiteLLMLogger):
    def __init__(
        self,
        api_key: Optional[str] = None,
        mut_name: Optional[str] = None,
        context_token: Optional[str] = None,
        tags: Optional[List[str]] = None,
        host_address: Optional[str] = None,
    ) -> None:
        if not api_key:
            api_key = os.getenv("OKAREO_API_KEY", "")
        if not mut_name:
            mut_name = os.getenv("OKAREO_MUT_NAME", "")
        if not tags:
            tags_str = os.getenv("OKAREO_MUT_TAGS", "")
            if tags_str and len(tags_str) > 0:
                tags = tags_str.split(",")
        if not context_token:
            context_token = os.getenv("OKAREO_CONTEXT_TOKEN", "")

        super().__init__(
            api_key=api_key,
            mut_name=mut_name,
            context_token=context_token,
            tags=tags,
            host_address=host_address,
        )


litellm_proxy_handler = LiteLLMProxyLogger()



from enum import Enum


class ScenarioType(str, Enum):
    COMMON_CONTRACTIONS = "COMMON_CONTRACTIONS"
    COMMON_MISSPELLINGS = "COMMON_MISSPELLINGS"
    CONDITIONAL = "CONDITIONAL"
    LABEL_REVERSE_INVARIANT = "LABEL_REVERSE_INVARIANT"
    NAMED_ENTITY_SUBSTITUTION = "NAMED_ENTITY_SUBSTITUTION"
    NEGATION = "NEGATION"
    REPHRASE_INVARIANT = "REPHRASE_INVARIANT"
    ROUNDTRIP_INVARIANT = "ROUNDTRIP_INVARIANT"
    SEED = "SEED"
    TERM_RELEVANCE_INVARIANT = "TERM_RELEVANCE_INVARIANT"
    TEXT_REVERSE_LABELED = "TEXT_REVERSE_LABELED"
    TEXT_REVERSE_QUESTION = "TEXT_REVERSE_QUESTION"
    WORD_CHARACTER_INSERTION = "WORD_CHARACTER_INSERTION"
    WORD_INFLECTIONS = "WORD_INFLECTIONS"
    WORD_QWERTY_MISSPELL = "WORD_QWERTY_MISSPELL"
    WORD_SYNONYM_EMBEDDING = "WORD_SYNONYM_EMBEDDING"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="FindTestDataPointPayload")


@_attrs_define
class FindTestDataPointPayload:
    """
    Attributes:
        id (Union[Unset, str]): ID of the datapoint
        test_run_id (Union[Unset, str]): ID of the test run
        scenario_data_point_id (Union[Unset, str]): ID of the scenario data point
        metric_type (Union[Unset, str]):
    """

    id: Union[Unset, str] = UNSET
    test_run_id: Union[Unset, str] = UNSET
    scenario_data_point_id: Union[Unset, str] = UNSET
    metric_type: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        test_run_id = self.test_run_id
        scenario_data_point_id = self.scenario_data_point_id
        metric_type = self.metric_type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if id is not UNSET:
            field_dict["id"] = id
        if test_run_id is not UNSET:
            field_dict["test_run_id"] = test_run_id
        if scenario_data_point_id is not UNSET:
            field_dict["scenario_data_point_id"] = scenario_data_point_id
        if metric_type is not UNSET:
            field_dict["metric_type"] = metric_type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        id = d.pop("id", UNSET)

        test_run_id = d.pop("test_run_id", UNSET)

        scenario_data_point_id = d.pop("scenario_data_point_id", UNSET)

        metric_type = d.pop("metric_type", UNSET)

        find_test_data_point_payload = cls(
            id=id,
            test_run_id=test_run_id,
            scenario_data_point_id=scenario_data_point_id,
            metric_type=metric_type,
        )

        find_test_data_point_payload.additional_properties = d
        return find_test_data_point_payload

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.generation_tone import GenerationTone
from ..models.scenario_type import ScenarioType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.seed_data import SeedData


T = TypeVar("T", bound="ScenarioSetCreate")


@_attrs_define
class ScenarioSetCreate:
    """
    Attributes:
        name (str): Name of the scenario set
        seed_data (List['SeedData']): Seed data is a list of dictionaries, each with an input and result
        number_examples (int): Number of examples
        project_id (Union[Unset, str]): ID for the project
        generation_type (Union[Unset, ScenarioType]): An enumeration. Default: ScenarioType.SEED.
        generation_tone (Union[Unset, GenerationTone]): An enumeration. Default: GenerationTone.NEUTRAL.
    """

    name: str
    seed_data: List["SeedData"]
    number_examples: int
    project_id: Union[Unset, str] = UNSET
    generation_type: Union[Unset, ScenarioType] = ScenarioType.SEED
    generation_tone: Union[Unset, GenerationTone] = GenerationTone.NEUTRAL
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        seed_data = []
        for seed_data_item_data in self.seed_data:
            seed_data_item = seed_data_item_data.to_dict()

            seed_data.append(seed_data_item)

        number_examples = self.number_examples
        project_id = self.project_id
        generation_type: Union[Unset, str] = UNSET
        if not isinstance(self.generation_type, Unset):
            generation_type = self.generation_type.value

        generation_tone: Union[Unset, str] = UNSET
        if not isinstance(self.generation_tone, Unset):
            generation_tone = self.generation_tone.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "seed_data": seed_data,
                "number_examples": number_examples,
            }
        )
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if generation_type is not UNSET:
            field_dict["generation_type"] = generation_type
        if generation_tone is not UNSET:
            field_dict["generation_tone"] = generation_tone

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.seed_data import SeedData

        d = src_dict.copy()
        name = d.pop("name")

        seed_data = []
        _seed_data = d.pop("seed_data")
        for seed_data_item_data in _seed_data:
            seed_data_item = SeedData.from_dict(seed_data_item_data)

            seed_data.append(seed_data_item)

        number_examples = d.pop("number_examples")

        project_id = d.pop("project_id", UNSET)

        _generation_type = d.pop("generation_type", UNSET)
        generation_type: Union[Unset, ScenarioType]
        if isinstance(_generation_type, Unset):
            generation_type = UNSET
        else:
            generation_type = ScenarioType(_generation_type)

        _generation_tone = d.pop("generation_tone", UNSET)
        generation_tone: Union[Unset, GenerationTone]
        if isinstance(_generation_tone, Unset):
            generation_tone = UNSET
        else:
            generation_tone = GenerationTone(_generation_tone)

        scenario_set_create = cls(
            name=name,
            seed_data=seed_data,
            number_examples=number_examples,
            project_id=project_id,
            generation_type=generation_type,
            generation_tone=generation_tone,
        )

        scenario_set_create.additional_properties = d
        return scenario_set_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os
from datetime import datetime
from typing import Any

import pytest
from okareo_tests.common import API_KEY

from okareo import Okareo
from okareo.model_under_test import CustomModel
from okareo_api_client.models import ScenarioSetResponse, ScenarioType, TestRunType

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


@pytest.fixture(scope="module")
def uploaded_scenario_set(okareo_client: Okareo) -> ScenarioSetResponse:
    file_path = os.path.join(os.path.dirname(__file__), "webbizz_3_test_article.jsonl")
    articles: ScenarioSetResponse = okareo_client.upload_scenario_set(
        file_path=file_path, scenario_name=f"test_upload_scenario_set {today_with_time}"
    )

    return articles


@pytest.fixture(scope="module")
def generate_scenarios(
    okareo_client: Okareo, uploaded_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    questions: ScenarioSetResponse = okareo_client.generate_scenarios(
        source_scenario=uploaded_scenario_set.scenario_id,
        name=f"test_generate_scenarios {today_with_time}",
        number_examples=2,
        generation_type=ScenarioType.TEXT_REVERSE_QUESTION,
    )
    return questions


def test_upload_scenario_set(uploaded_scenario_set: ScenarioSetResponse) -> None:
    assert uploaded_scenario_set is not None
    assert uploaded_scenario_set.scenario_id
    assert uploaded_scenario_set.project_id
    assert uploaded_scenario_set.time_created


def test_generate_scenarios(generate_scenarios: ScenarioSetResponse) -> None:
    assert generate_scenarios is not None
    assert generate_scenarios.scenario_id
    assert generate_scenarios.project_id
    assert generate_scenarios.time_created


def test_run_test_retrieval(
    okareo_client: Okareo, generate_scenarios: ScenarioSetResponse
) -> None:
    class RetrievalModel(CustomModel):
        # Callable to be applied to each scenario in the scenario set
        def invoke(self, input_value: str) -> Any:
            # call your embedding model and vector db retrieval being tested here using <input> from the scenario set
            # we are using a random response here for demonstration purposes
            article_ids = [
                "75eaa363-dfcc-499f-b2af-1407b43cb133",
                "ac0d464c-f673-44b8-8195-60c965e47525",
            ]
            import random

            selected_ids = random.sample(article_ids, 1)
            rounded_random_scores = sorted(
                [round(random.random(), 2) for _ in range(5)], reverse=True
            )

            # higher score value means more relevant
            parsed_ids_with_scores = [
                (selected_id, score)
                for selected_id, score in zip(selected_ids, rounded_random_scores)
            ]

            model_response = {"matches": "additional context from the model"}

            # return a tuple of (parsed_ids_with_scores, overall model response context)
            return parsed_ids_with_scores, model_response

    model_under_test = okareo_client.register_model(
        name="vectordb_retrieval test",
        model=RetrievalModel(name="custom retrieval"),
    )

    test_run_name = f"test_run_test_retrieval {today_with_time}"

    retrieval_test_run = model_under_test.run_test(
        scenario=generate_scenarios,
        name=test_run_name,
        test_run_type=TestRunType.INFORMATION_RETRIEVAL,
    )

    assert retrieval_test_run.id
    assert retrieval_test_run.model_metrics



import inspect
import json
from abc import abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from attrs import define as _attrs_define

from okareo.error import MissingApiKeyError, MissingVectorDbError
from okareo_api_client.api.default import (
    add_datapoint_v0_datapoints_post,
    get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get,
    get_test_run_v0_test_runs_test_run_id_get,
    run_test_v0_test_run_post,
)
from okareo_api_client.client import Client
from okareo_api_client.errors import UnexpectedStatus
from okareo_api_client.models import (
    DatapointResponse,
    DatapointSchema,
    ModelUnderTestResponse,
    ScenarioDataPoinResponse,
    TestRunItem,
    TestRunType,
)
from okareo_api_client.models.error_response import ErrorResponse
from okareo_api_client.models.scenario_set_response import ScenarioSetResponse
from okareo_api_client.models.test_run_payload_v2 import TestRunPayloadV2
from okareo_api_client.models.test_run_payload_v2_api_keys import (
    TestRunPayloadV2ApiKeys,
)
from okareo_api_client.models.test_run_payload_v2_metrics_kwargs import (
    TestRunPayloadV2MetricsKwargs,
)
from okareo_api_client.models.test_run_payload_v2_model_results import (
    TestRunPayloadV2ModelResults,
)
from okareo_api_client.types import UNSET, Unset

from .async_utils import AsyncProcessorMixin


class BaseModel:
    type: str
    api_key: Optional[str]

    @abstractmethod
    def params(self) -> dict:
        pass


@_attrs_define
class OpenAIModel(BaseModel):
    type = "openai"
    model_id: str
    temperature: float
    system_prompt_template: str
    user_prompt_template: Optional[str] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "temperature": self.temperature,
            "system_prompt_template": self.system_prompt_template,
            "user_prompt_template": self.user_prompt_template,
        }


@_attrs_define
class CohereModel(BaseModel):
    type = "cohere"
    model_id: str
    model_type: str
    input_type: Optional[str] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "model_type": self.model_type,
            "input_type": self.input_type,
        }


@_attrs_define
class PineconeDb(BaseModel):
    type = "pinecone"
    index_name: str
    region: str
    project_id: str
    top_k: int = 5

    def params(self) -> dict:
        return {
            "index_name": self.index_name,
            "region": self.region,
            "project_id": self.project_id,
            "top_k": self.top_k,
        }


@_attrs_define
class QdrantDB(BaseModel):
    type = "qdrant"
    collection_name: str
    url: str
    top_k: int = 5
    sparse: bool = False

    def params(self) -> dict:
        return {
            "collection_name": self.collection_name,
            "url": self.url,
            "top_k": self.top_k,
            "sparse": self.sparse,
        }


@_attrs_define
class CustomModel(BaseModel):
    type = "custom"
    name: str

    @abstractmethod
    def invoke(self, input_value: str) -> Any:
        pass

    def params(self) -> dict:
        return {
            "name": self.name,
            "type": self.type,
            "model_invoker": self.invoke,
        }


class ModelUnderTest(AsyncProcessorMixin):
    def __init__(
        self,
        client: Client,
        api_key: str,
        mut: ModelUnderTestResponse,
        models: Optional[Dict[str, Any]] = None,
    ):
        self.client = client
        self.api_key = api_key

        self.mut_id = mut.id
        self.project_id = mut.project_id
        self.name = mut.name
        self.tags = mut.tags
        self.models = models
        self.app_link = mut.app_link
        super().__init__(name="OkareoDatapointsProcessor")

    def get_client(self) -> Client:
        return self.client

    def get_api_key(self) -> str:
        return self.api_key

    def add_data_point(
        self,
        input_obj: Union[dict, str, None] = None,
        result_obj: Union[dict, str, None] = None,
        feedback: Union[float, None] = None,
        context_token: Union[str, None] = None,
        error_message: Union[str, None] = None,
        error_code: Union[str, None] = None,
        input_datetime: Union[str, Unset] = UNSET,
        result_datetime: Union[str, Unset] = UNSET,
        project_id: Union[str, None] = None,
        tags: Union[List[str], None] = None,
        test_run_id: Union[None, str] = None,
    ) -> Union[DatapointResponse, ErrorResponse]:
        body = {
            "tags": tags or [],
            "input": json.dumps(input_obj, default=str),
            "result": json.dumps(result_obj, default=str),
            "context_token": context_token,
            "feedback": feedback,
            "error_message": error_message,
            "error_code": error_code,
            "input_datetime": (
                datetime.now().isoformat()
                if input_datetime == UNSET and input_obj is not None
                else input_datetime
            ),
            "result_datetime": (
                datetime.now().isoformat()
                if result_datetime == UNSET and result_obj is not None
                else result_datetime
            ),
            "project_id": self.project_id,
            "mut_id": self.mut_id,
            "test_run_id": test_run_id,
        }
        response = add_datapoint_v0_datapoints_post.sync(
            client=self.client,
            api_key=self.api_key,
            json_body=DatapointSchema.from_dict(body),
        )
        if not response:
            print("Empty response from API")
        assert response is not None

        return response

    def add_data_point_async(
        self,
        input_obj: Union[dict, str, None] = None,
        result_obj: Union[dict, str, None] = None,
        feedback: Union[float, None] = None,
        context_token: Union[str, None] = None,
        error_message: Union[str, None] = None,
        error_code: Union[str, None] = None,
        input_datetime: Union[str, Unset] = UNSET,
        result_datetime: Union[str, Unset] = UNSET,
        project_id: Union[str, None] = None,
        tags: Union[List[str], None] = None,
        test_run_id: Union[None, str] = None,
    ) -> bool:
        body = {
            "tags": tags or [],
            "input": json.dumps(input_obj, default=str),
            "result": json.dumps(result_obj, default=str),
            "context_token": context_token,
            "feedback": feedback,
            "error_message": error_message,
            "error_code": error_code,
            "input_datetime": (
                datetime.now().isoformat()
                if input_datetime == UNSET and input_obj is not None
                else input_datetime
            ),
            "result_datetime": (
                datetime.now().isoformat()
                if result_datetime == UNSET and result_obj is not None
                else result_datetime
            ),
            "project_id": self.project_id,
            "mut_id": self.mut_id,
            "test_run_id": test_run_id,
        }

        return self.async_call(
            add_datapoint_v0_datapoints_post.sync, DatapointSchema.from_dict(body)
        )

    def _validate_run_test_params(
        self,
        api_key: Optional[str],
        api_keys: Optional[dict],
        test_run_type: TestRunType,
    ) -> dict:
        assert isinstance(self.models, dict)
        model_names = list(self.models.keys())
        run_api_keys = api_keys if api_keys else {model_names[0]: api_key}

        if "custom" not in model_names and len(model_names) != len(run_api_keys):
            raise MissingApiKeyError("Number of models and API keys does not match")

        if test_run_type == TestRunType.INFORMATION_RETRIEVAL:
            if {"pinecone", "qdrant", "custom"}.isdisjoint(model_names):
                raise MissingVectorDbError("No vector database specified")

        return run_api_keys

    def _has_custom_model(self) -> bool:
        assert isinstance(self.models, dict)
        return "custom" in list(self.models.keys())

    def _get_scenario_data_points(
        self, scenario_id: str
    ) -> List[ScenarioDataPoinResponse]:
        scenario_data_points = (
            get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get.sync(
                client=self.client,
                api_key=self.api_key,
                scenario_id=scenario_id,
            )
        )
        scenario_data_points = (
            scenario_data_points if isinstance(scenario_data_points, List) else []
        )
        return scenario_data_points

    def _get_test_run_payload(
        self,
        scenario_id: str,
        name: str,
        api_key: Optional[str],
        api_keys: Optional[dict],
        run_api_keys: dict,
        metrics_kwargs: Optional[dict],
        test_run_type: TestRunType,
        calculate_metrics: bool,
        model_data: dict,
        checks: Optional[List[str]],
    ) -> TestRunPayloadV2:
        return TestRunPayloadV2(
            mut_id=self.mut_id,
            api_keys=(
                TestRunPayloadV2ApiKeys.from_dict(run_api_keys)
                if api_keys or api_key
                else UNSET
            ),
            scenario_id=scenario_id,
            name=name,
            type=test_run_type,
            project_id=self.project_id,
            calculate_metrics=calculate_metrics,
            metrics_kwargs=TestRunPayloadV2MetricsKwargs.from_dict(
                metrics_kwargs or {}
            ),
            model_results=(
                TestRunPayloadV2ModelResults.from_dict(model_data)
                if self._has_custom_model()
                else UNSET
            ),
            checks=checks if checks else UNSET,
        )

    def run_test(
        self,
        scenario: Union[ScenarioSetResponse, str],
        name: str,
        api_key: Optional[str] = None,
        api_keys: Optional[dict] = None,
        metrics_kwargs: Optional[dict] = None,
        test_run_type: TestRunType = TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics: bool = False,
        checks: Optional[List[str]] = None,
    ) -> TestRunItem:
        """Server-based version of test-run execution"""
        try:
            assert isinstance(self.models, dict)
            scenario_id = (
                scenario.scenario_id
                if isinstance(scenario, ScenarioSetResponse)
                else scenario
            )
            run_api_keys = self._validate_run_test_params(
                api_key, api_keys, test_run_type
            )

            model_data: dict = {"model_data": {}}
            if self._has_custom_model():
                scenario_data_points = self._get_scenario_data_points(scenario_id)

                custom_model_invoker = self.models["custom"]["model_invoker"]
                for scenario_data_point in scenario_data_points:
                    assert isinstance(scenario_data_point.input_, str)
                    actual, model_response = custom_model_invoker(
                        scenario_data_point.input_
                    )
                    model_data["model_data"][scenario_data_point.id] = {
                        "actual": actual,
                        "model_response": model_response,
                    }

            response = run_test_v0_test_run_post.sync(
                client=self.client,
                api_key=self.api_key,
                json_body=self._get_test_run_payload(
                    scenario_id,
                    name,
                    api_key,
                    api_keys,
                    run_api_keys,
                    metrics_kwargs,
                    test_run_type,
                    calculate_metrics,
                    model_data,
                    checks,
                ),
            )
        except UnexpectedStatus as e:
            print(f"Unexpected status {e=}, {e.content=}")
            raise

        if isinstance(response, ErrorResponse):
            error_message = f"error: {response}, {response.detail}"
            print(error_message)
            raise
        if not response:
            print("Empty response from API")
        assert response is not None
        return response

    async def run_test_async(
        self,
        scenario: Union[ScenarioSetResponse, str],
        name: str,
        api_key: Optional[str] = None,
        api_keys: Optional[dict] = None,
        metrics_kwargs: Optional[dict] = None,
        test_run_type: TestRunType = TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics: bool = False,
        checks: Optional[List[str]] = None,
    ) -> TestRunItem:
        """Server-based version of test-run execution"""
        try:
            assert isinstance(self.models, dict)
            scenario_id = (
                scenario.scenario_id
                if isinstance(scenario, ScenarioSetResponse)
                else scenario
            )
            run_api_keys = self._validate_run_test_params(
                api_key, api_keys, test_run_type
            )

            model_data: dict = {"model_data": {}}
            if self._has_custom_model():
                scenario_data_points = self._get_scenario_data_points(scenario_id)

                custom_model_invoker = self.models["custom"]["model_invoker"]
                async_custom_model = inspect.iscoroutinefunction(custom_model_invoker)
                for scenario_data_point in scenario_data_points:
                    assert isinstance(scenario_data_point.input_, str)
                    if async_custom_model:
                        # If the method is async, await it
                        actual, model_response = await custom_model_invoker(
                            scenario_data_point.input_
                        )
                    else:
                        # If the method is not async, call it directly
                        actual, model_response = custom_model_invoker(
                            scenario_data_point.input_
                        )
                    model_data["model_data"][scenario_data_point.id] = {
                        "actual": actual,
                        "model_response": model_response,
                    }

            response = run_test_v0_test_run_post.sync(
                client=self.client,
                api_key=self.api_key,
                json_body=self._get_test_run_payload(
                    scenario_id,
                    name,
                    api_key,
                    api_keys,
                    run_api_keys,
                    metrics_kwargs,
                    test_run_type,
                    calculate_metrics,
                    model_data,
                    checks,
                ),
            )
        except UnexpectedStatus as e:
            print(f"Unexpected status {e=}, {e.content=}")
            raise

        if isinstance(response, ErrorResponse):
            error_message = f"error: {response}, {response.detail}"
            print(error_message)
            raise
        if not response:
            print("Empty response from API")
        assert response is not None
        return response

    def get_test_run(self, test_run_id: str) -> TestRunItem:
        try:
            response = get_test_run_v0_test_runs_test_run_id_get.sync(
                client=self.client, api_key=self.api_key, test_run_id=test_run_id
            )
            if not isinstance(response, TestRunItem):
                raise
            return response
        except UnexpectedStatus as e:
            print(e.content)
            raise



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="SummaryDatapointSearch")


@_attrs_define
class SummaryDatapointSearch:
    """
    Attributes:
        project_id (Union[Unset, str]): Project ID
        mut_id (Union[Unset, str]): Model ID
        from_date (Union[Unset, datetime.datetime]): Earliest date Default: isoparse('2022-12-31T23:59:59.999999').
        to_date (Union[Unset, datetime.datetime]): Latest date
    """

    project_id: Union[Unset, str] = UNSET
    mut_id: Union[Unset, str] = UNSET
    from_date: Union[Unset, datetime.datetime] = isoparse("2022-12-31T23:59:59.999999")
    to_date: Union[Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        project_id = self.project_id
        mut_id = self.mut_id
        from_date: Union[Unset, str] = UNSET
        if not isinstance(self.from_date, Unset):
            from_date = self.from_date.isoformat()

        to_date: Union[Unset, str] = UNSET
        if not isinstance(self.to_date, Unset):
            to_date = self.to_date.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if from_date is not UNSET:
            field_dict["from_date"] = from_date
        if to_date is not UNSET:
            field_dict["to_date"] = to_date

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        project_id = d.pop("project_id", UNSET)

        mut_id = d.pop("mut_id", UNSET)

        _from_date = d.pop("from_date", UNSET)
        from_date: Union[Unset, datetime.datetime]
        if isinstance(_from_date, Unset):
            from_date = UNSET
        else:
            from_date = isoparse(_from_date)

        _to_date = d.pop("to_date", UNSET)
        to_date: Union[Unset, datetime.datetime]
        if isinstance(_to_date, Unset):
            to_date = UNSET
        else:
            to_date = isoparse(_to_date)

        summary_datapoint_search = cls(
            project_id=project_id,
            mut_id=mut_id,
            from_date=from_date,
            to_date=to_date,
        )

        summary_datapoint_search.additional_properties = d
        return summary_datapoint_search

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Dict, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.error_response import ErrorResponse
from ...models.project_response import ProjectResponse
from ...models.project_schema import ProjectSchema
from ...types import Response


def _get_kwargs(
    *,
    json_body: ProjectSchema,
    api_key: str,
) -> Dict[str, Any]:
    headers = {}
    headers["api-key"] = api_key

    json_json_body = json_body.to_dict()

    return {
        "method": "post",
        "url": "/v0/projects",
        "json": json_json_body,
        "headers": headers,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[ErrorResponse, ProjectResponse]]:
    if response.status_code == HTTPStatus.CREATED:
        response_201 = ProjectResponse.from_dict(response.json())

        return response_201
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[ErrorResponse, ProjectResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: ProjectSchema,
    api_key: str,
) -> Response[Union[ErrorResponse, ProjectResponse]]:
    """Create Project

     Create a project

    Returns:
        the id, name, and tags for the created project

    Args:
        api_key (str):
        json_body (ProjectSchema):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, ProjectResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
        api_key=api_key,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: ProjectSchema,
    api_key: str,
) -> Optional[Union[ErrorResponse, ProjectResponse]]:
    """Create Project

     Create a project

    Returns:
        the id, name, and tags for the created project

    Args:
        api_key (str):
        json_body (ProjectSchema):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, ProjectResponse]
    """

    return sync_detailed(
        client=client,
        json_body=json_body,
        api_key=api_key,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: ProjectSchema,
    api_key: str,
) -> Response[Union[ErrorResponse, ProjectResponse]]:
    """Create Project

     Create a project

    Returns:
        the id, name, and tags for the created project

    Args:
        api_key (str):
        json_body (ProjectSchema):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, ProjectResponse]]
    """

    kwargs = _get_kwargs(
        json_body=json_body,
        api_key=api_key,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    json_body: ProjectSchema,
    api_key: str,
) -> Optional[Union[ErrorResponse, ProjectResponse]]:
    """Create Project

     Create a project

    Returns:
        the id, name, and tags for the created project

    Args:
        api_key (str):
        json_body (ProjectSchema):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, ProjectResponse]
    """

    return (
        await asyncio_detailed(
            client=client,
            json_body=json_body,
            api_key=api_key,
        )
    ).parsed



import os
from typing import Any, Dict, List, Union

import httpx

from okareo_api_client import Client
from okareo_api_client.api.default import (
    create_scenario_set_v0_scenario_sets_post,
    evaluator_generate_v0_evaluator_generate_post,
    evaluator_upload_v0_evaluator_upload_post,
    generate_scenario_set_v0_scenario_sets_generate_post,
    get_all_projects_v0_projects_get,
    get_datapoints_v0_find_datapoints_post,
    get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get,
    register_model_v0_register_model_post,
    scenario_sets_upload_v0_scenario_sets_upload_post,
)
from okareo_api_client.errors import UnexpectedStatus
from okareo_api_client.models.body_evaluator_upload_v0_evaluator_upload_post import (
    BodyEvaluatorUploadV0EvaluatorUploadPost,
)
from okareo_api_client.models.body_scenario_sets_upload_v0_scenario_sets_upload_post import (
    BodyScenarioSetsUploadV0ScenarioSetsUploadPost,
)
from okareo_api_client.models.datapoint_list_item import DatapointListItem
from okareo_api_client.models.datapoint_search import DatapointSearch
from okareo_api_client.models.error_response import ErrorResponse
from okareo_api_client.models.evaluator_detailed_response import (
    EvaluatorDetailedResponse,
)
from okareo_api_client.models.evaluator_generate_response import (
    EvaluatorGenerateResponse,
)
from okareo_api_client.models.evaluator_spec_request import EvaluatorSpecRequest
from okareo_api_client.models.model_under_test_response import ModelUnderTestResponse
from okareo_api_client.models.model_under_test_schema import ModelUnderTestSchema
from okareo_api_client.models.project_response import ProjectResponse
from okareo_api_client.models.scenario_data_poin_response import (
    ScenarioDataPoinResponse,
)
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.scenario_set_generate import ScenarioSetGenerate
from okareo_api_client.models.scenario_set_response import ScenarioSetResponse
from okareo_api_client.models.scenario_type import ScenarioType
from okareo_api_client.types import UNSET, File, Unset

from .common import BASE_URL, HTTPX_TIME_OUT
from .model_under_test import BaseModel, ModelUnderTest


class Okareo:
    """A class for interacting with Okareo API"""

    def __init__(
        self, api_key: str, base_path: str = BASE_URL, timeout: float = HTTPX_TIME_OUT
    ):
        self.api_key = api_key
        self.client = Client(
            base_url=base_path, raise_on_unexpected_status=True
        )  # otherwise everything except 201 and 422 is swallowed

    def get_projects(self) -> List[ProjectResponse]:
        response = get_all_projects_v0_projects_get.sync(
            client=self.client,
            api_key=self.api_key,
        )
        self.validate_response(response)
        assert isinstance(response, List)

        return response

    def register_model(
        self,
        name: str,
        tags: Union[List[str], None] = None,
        project_id: Union[str, None] = None,
        model: Union[None, BaseModel, List[BaseModel]] = None,
    ) -> ModelUnderTest:
        if tags is None:
            tags = []
        data: Dict[str, Any] = {"name": name, "tags": tags}
        # will rename name to model in the future api-breaking release
        model_invoker = None
        if isinstance(model, BaseModel) or (
            isinstance(model, list) and all(isinstance(x, BaseModel) for x in model)
        ):
            models = model if isinstance(model, list) else [model]
            data["models"] = {}
            for model in models:
                data["models"][model.type] = model.params()
            if "custom" in data["models"].keys():
                model_invoker = data["models"]["custom"]["model_invoker"]
                del data["models"]["custom"]["model_invoker"]
        if project_id is not None:
            data["project_id"] = project_id
        json_body = ModelUnderTestSchema.from_dict(data)
        response = register_model_v0_register_model_post.sync(
            client=self.client, api_key=self.api_key, json_body=json_body
        )

        self.validate_response(response)
        assert isinstance(response, ModelUnderTestResponse)
        model_data = data.get("models")
        if model_invoker and isinstance(model_data, dict):
            model_data["custom"]["model_invoker"] = model_invoker
        return ModelUnderTest(
            client=self.client,
            api_key=self.api_key,
            mut=response,
            models=model_data,
        )

    def create_scenario_set(
        self, create_request: ScenarioSetCreate
    ) -> ScenarioSetResponse:
        response = create_scenario_set_v0_scenario_sets_post.sync(
            client=self.client, api_key=self.api_key, json_body=create_request
        )

        self.validate_response(response)
        assert isinstance(response, ScenarioSetResponse)

        return response

    def upload_scenario_set(
        self, scenario_name: str, file_path: str, project_id: Union[Unset, str] = UNSET
    ) -> ScenarioSetResponse:
        try:
            file_name = os.path.basename(file_path)

            with open(file_path, "rb") as binary_io:
                multipart_body = BodyScenarioSetsUploadV0ScenarioSetsUploadPost(
                    name=scenario_name,
                    project_id=project_id,
                    file=File(file_name=file_name, payload=binary_io),
                )
                response = scenario_sets_upload_v0_scenario_sets_upload_post.sync(
                    client=self.client,
                    api_key=self.api_key,
                    multipart_data=multipart_body,
                )

            self.validate_response(response)
            assert isinstance(response, ScenarioSetResponse)

            return response
        except UnexpectedStatus as e:
            print(e.content)
            raise

    def download_scenario_set(
        self,
        scenario: Union[ScenarioSetResponse, str],
        file_path: str = "",
    ) -> Any:
        try:
            scenario_id = (
                scenario if isinstance(scenario, str) else scenario.scenario_id
            )
            url = f"{BASE_URL}/v0/scenario_sets_download/{scenario_id}"
            headers = {
                "accept": "application/json",
                "api-key": self.api_key,
            }
            response = httpx.get(
                url,
                headers=headers,
            )
            filename = response.headers["content-disposition"].split('"')[1]
            if file_path != "":
                filename = file_path
            binary_file = open(filename, "wb")
            binary_file.write(response.content)
            binary_file.close()
            return binary_file
        except Exception as e:
            print(e)
            raise

    def generate_scenarios(
        self,
        source_scenario: Union[str, ScenarioSetResponse],
        name: str,
        number_examples: int,
        project_id: Union[Unset, str] = UNSET,
        generation_type: Union[Unset, ScenarioType] = ScenarioType.REPHRASE_INVARIANT,
    ) -> ScenarioSetResponse:
        scenario_id = (
            source_scenario.scenario_id
            if isinstance(source_scenario, ScenarioSetResponse)
            else source_scenario
        )
        return self.generate_scenario_set(
            ScenarioSetGenerate(
                source_scenario_id=scenario_id,
                name=name,
                number_examples=number_examples,
                project_id=project_id,
                generation_type=generation_type,
            )
        )

    def generate_scenario_set(
        self, create_request: ScenarioSetGenerate
    ) -> ScenarioSetResponse:
        response = generate_scenario_set_v0_scenario_sets_generate_post.sync(
            client=self.client, api_key=self.api_key, json_body=create_request
        )

        self.validate_response(response)
        assert isinstance(response, ScenarioSetResponse)

        return response

    def get_scenario_data_points(
        self, scenario_id: str
    ) -> List[ScenarioDataPoinResponse]:
        response = (
            get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get.sync(
                client=self.client, api_key=self.api_key, scenario_id=scenario_id
            )
        )

        self.validate_response(response)
        assert isinstance(response, List)

        return response

    def validate_response(self, response: Any) -> None:
        if isinstance(response, ErrorResponse):
            error_message = f"error: {response}, {response.detail}"
            print(error_message)
            raise TypeError(error_message)
        if response is None:
            print("Received no response (None) from the API")
            raise ValueError("No response received")

    def find_datapoints(
        self, context_token: str
    ) -> Union[List[DatapointListItem], ErrorResponse]:
        data = get_datapoints_v0_find_datapoints_post.sync(
            client=self.client,
            api_key=self.api_key,
            json_body=DatapointSearch(context_token=context_token),
        )
        if not data:
            return []
        return data

    def upload_evaluator(
        self,
        name: str,
        file_path: str,
        requires_scenario_input: bool,
        requires_scenario_result: bool,
        description: str = "",
        output_data_type: str = "",
        project_id: Union[Unset, str] = UNSET,
    ) -> EvaluatorDetailedResponse:
        try:
            file_name = os.path.basename(file_path)

            with open(file_path, "rb") as binary_io:
                multipart_body = BodyEvaluatorUploadV0EvaluatorUploadPost(
                    name=name,
                    requires_scenario_input=requires_scenario_input,
                    requires_scenario_result=requires_scenario_result,
                    description=description,
                    output_data_type=output_data_type,
                    project_id=project_id,
                    file=File(file_name=file_name, payload=binary_io),
                )
                response = evaluator_upload_v0_evaluator_upload_post.sync(
                    client=self.client,
                    multipart_data=multipart_body,
                    api_key=self.api_key,
                )

            self.validate_response(response)
            assert isinstance(response, EvaluatorDetailedResponse)

            return response
        except UnexpectedStatus as e:
            print(e.content)
            raise

    def generate_evaluator(
        self, create_evaluator: EvaluatorSpecRequest
    ) -> EvaluatorGenerateResponse:
        response = evaluator_generate_v0_evaluator_generate_post.sync(
            client=self.client, api_key=self.api_key, json_body=create_evaluator
        )
        self.validate_response(response)
        assert isinstance(response, EvaluatorGenerateResponse)

        return response



from enum import Enum


class TestRunType(str, Enum):
    INFORMATION_RETRIEVAL = "INFORMATION_RETRIEVAL"
    INVARIANT = "invariant"
    MULTI_CLASS_CLASSIFICATION = "MULTI_CLASS_CLASSIFICATION"
    NL_GENERATION = "NL_GENERATION"

    def __str__(self) -> str:
        return str(self.value)



import json
import os
import time
from datetime import datetime
from typing import List

import pytest
import requests  # type: ignore
from okareo_tests.common import API_KEY

from okareo import Okareo
from okareo_api_client.models import (
    ScenarioSetCreate,
    ScenarioSetGenerate,
    ScenarioSetResponse,
    ScenarioType,
    SeedData,
)
from okareo_api_client.models.scenario_data_poin_response import (
    ScenarioDataPoinResponse,
)

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


@pytest.fixture(scope="module")
def seed_data() -> List[SeedData]:
    return [
        SeedData(input_="example question or statement", result="example result"),
        SeedData(
            input_="tell me about your capability", result="unique example result"
        ),
        SeedData(input_="what are your limitations", result="different example result"),
    ]


@pytest.fixture(scope="module")
def create_scenario_set(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name="my test scenario set",
        number_examples=1,
        seed_data=seed_data,
    )
    articles: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    return articles


@pytest.fixture(scope="module")
def generate_scenarios(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> ScenarioSetResponse:
    scenario_set_generate = ScenarioSetGenerate(
        source_scenario_id=create_scenario_set.scenario_id,
        name="generated scenario set",
        number_examples=2,
    )
    response = okareo_client.generate_scenario_set(scenario_set_generate)
    return response


@pytest.fixture(scope="module")
def get_scenario_data_points(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> List[ScenarioDataPoinResponse]:
    return okareo_client.get_scenario_data_points(create_scenario_set.scenario_id)


def test_create_scenario_set(
    create_scenario_set: ScenarioSetResponse, seed_data: List[SeedData]
) -> None:
    assert create_scenario_set.type == "SEED"  # default test
    assert create_scenario_set.scenario_id
    assert create_scenario_set.project_id
    assert create_scenario_set.time_created
    if isinstance(create_scenario_set.seed_data, List):
        for i in range(3):
            assert seed_data[i].input_ == create_scenario_set.seed_data[i].input_
            assert seed_data[i].result == create_scenario_set.seed_data[i].result


def test_download_scenario_set(
    okareo_client: Okareo,
    create_scenario_set: ScenarioSetResponse,
    seed_data: List[SeedData],
) -> None:
    response_file = okareo_client.download_scenario_set(create_scenario_set)
    with open(response_file.name) as scenario_file:
        for line in scenario_file:
            assert json.loads(line)["input"] != ""
            assert json.loads(line)["result"] != ""


def test_download_scenario_set_with_file(
    okareo_client: Okareo,
    create_scenario_set: ScenarioSetResponse,
    seed_data: List[SeedData],
) -> None:
    response_file = okareo_client.download_scenario_set(
        create_scenario_set, "./scenario.jsonl"
    )
    with open(response_file.name) as scenario_file:
        for line in scenario_file:
            assert json.loads(line)["input"] != ""
            assert json.loads(line)["result"] != ""


def test_create_scenario_set_rephrase(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    scenario_set_create = ScenarioSetCreate(
        name="my test scenario set",
        generation_type=ScenarioType.REPHRASE_INVARIANT,
        number_examples=2,
        seed_data=seed_data,
    )
    create_scenario_set: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    assert create_scenario_set.type == "REPHRASE_INVARIANT"
    assert create_scenario_set.scenario_id
    assert create_scenario_set.project_id
    assert create_scenario_set.time_created
    assert isinstance(create_scenario_set.scenario_input, List)
    assert isinstance(create_scenario_set.seed_data, List)
    assert len(create_scenario_set.scenario_input) == 6
    for i in range(3):
        assert seed_data[i].input_ != create_scenario_set.scenario_input[i]


def test_generate_scenarios(
    generate_scenarios: ScenarioSetResponse, seed_data: List[SeedData]
) -> None:
    assert generate_scenarios.type == "REPHRASE_INVARIANT"
    assert generate_scenarios.seed_data == []
    assert generate_scenarios is not None
    assert generate_scenarios.scenario_id
    assert generate_scenarios.project_id
    assert generate_scenarios.time_created


def test_get_scenario_data_points(
    get_scenario_data_points: List[ScenarioDataPoinResponse], seed_data: List[SeedData]
) -> None:
    assert get_scenario_data_points is not None
    assert len(get_scenario_data_points) == 3


def delete_scenario_data_points(
    api_key: str, scenario_id: str, scenario_name: str
) -> None:
    url = f"https://api.okareo.com/v0/scenario_sets/{scenario_id}?name={scenario_name}"
    headers = {
        "Content-Type": "application/json",
        "accept": "application/json",
        "api-key": api_key,
    }
    requests.delete(url, headers=headers)


def dtest_create_delete_scenario_set_contraction_tiny_load(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    seed_data = []
    number_examples = 2
    file_path = os.path.join(
        os.path.dirname(__file__), "datasets/random_sentence_tiny.txt"
    )
    with open(file_path) as file:
        lines = file.readlines()
        for line in lines:
            parts = line.split("\t")
            assert len(parts) == 2
            seed_data.append(
                SeedData(input_=parts[0], result=parts[1]),
            )

    scenario_set_create = ScenarioSetCreate(
        name="my contraction test scenario set tiny load",
        generation_type=ScenarioType.COMMON_CONTRACTIONS,
        number_examples=number_examples,
        seed_data=seed_data,
    )
    scenario: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    time.sleep(10)

    if isinstance(scenario.tags, list) and isinstance(scenario.name, str):
        scenario_seed_id = scenario.tags[0].split(":")[1]
        delete_scenario_data_points(API_KEY, scenario_seed_id, scenario.name)
        delete_scenario_data_points(API_KEY, scenario.scenario_id, scenario.name)

    assert scenario.type == "COMMON_CONTRACTIONS"


def dtest_create_delete_scenario_set_misspelling_tiny_load(
    okareo_client: Okareo, seed_data: List[SeedData]
) -> None:
    seed_data = []
    number_examples = 2
    file_path = os.path.join(
        os.path.dirname(__file__), "datasets/random_sentence_tiny.txt"
    )
    with open(file_path) as file:
        lines = file.readlines()
        for line in lines:
            parts = line.split("\t")
            assert len(parts) == 2
            seed_data.append(
                SeedData(input_=parts[0], result=parts[1]),
            )

    scenario_set_create = ScenarioSetCreate(
        name="my misspelling test scenario set tiny load",
        generation_type=ScenarioType.COMMON_MISSPELLINGS,
        number_examples=number_examples,
        seed_data=seed_data,
    )
    scenario: ScenarioSetResponse = okareo_client.create_scenario_set(
        scenario_set_create
    )

    time.sleep(10)

    if isinstance(scenario.tags, list) and isinstance(scenario.name, str):
        scenario_seed_id = scenario.tags[0].split(":")[1]
        delete_scenario_data_points(API_KEY, scenario_seed_id, scenario.name)
        delete_scenario_data_points(API_KEY, scenario.scenario_id, scenario.name)

    assert scenario.type == "COMMON_MISSPELLINGS"



from typing import Any, Dict, List, Type, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="SemanticResult")


@_attrs_define
class SemanticResult:
    """
    Attributes:
        document_sentences (List[str]):
        summary_sentences (List[str]):
        scores (List[List[str]]):
    """

    document_sentences: List[str]
    summary_sentences: List[str]
    scores: List[List[str]]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        document_sentences = self.document_sentences

        summary_sentences = self.summary_sentences

        scores = []
        for scores_item_data in self.scores:
            scores_item = scores_item_data

            scores.append(scores_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "document_sentences": document_sentences,
                "summary_sentences": summary_sentences,
                "scores": scores,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        document_sentences = cast(List[str], d.pop("document_sentences"))

        summary_sentences = cast(List[str], d.pop("summary_sentences"))

        scores = []
        _scores = d.pop("scores")
        for scores_item_data in _scores:
            scores_item = cast(List[str], scores_item_data)

            scores.append(scores_item)

        semantic_result = cls(
            document_sentences=document_sentences,
            summary_sentences=summary_sentences,
            scores=scores,
        )

        semantic_result.additional_properties = d
        return semantic_result

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.feedback_range_summary import FeedbackRangeSummary


T = TypeVar("T", bound="DatapointSummaryItem")


@_attrs_define
class DatapointSummaryItem:
    """
    Attributes:
        date (datetime.datetime): The date for the data summary.
        feedback_ranges (Union[Unset, List['FeedbackRangeSummary']]): List of feedback range summaries for the date.
    """

    date: datetime.datetime
    feedback_ranges: Union[Unset, List["FeedbackRangeSummary"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        date = self.date.isoformat()

        feedback_ranges: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.feedback_ranges, Unset):
            feedback_ranges = []
            for feedback_ranges_item_data in self.feedback_ranges:
                feedback_ranges_item = feedback_ranges_item_data.to_dict()

                feedback_ranges.append(feedback_ranges_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "date": date,
            }
        )
        if feedback_ranges is not UNSET:
            field_dict["feedback_ranges"] = feedback_ranges

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.feedback_range_summary import FeedbackRangeSummary

        d = src_dict.copy()
        date = isoparse(d.pop("date"))

        feedback_ranges = []
        _feedback_ranges = d.pop("feedback_ranges", UNSET)
        for feedback_ranges_item_data in _feedback_ranges or []:
            feedback_ranges_item = FeedbackRangeSummary.from_dict(feedback_ranges_item_data)

            feedback_ranges.append(feedback_ranges_item)

        datapoint_summary_item = cls(
            date=date,
            feedback_ranges=feedback_ranges,
        )

        datapoint_summary_item.additional_properties = d
        return datapoint_summary_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="DatapointSearch")


@_attrs_define
class DatapointSearch:
    """
    Attributes:
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter datapoints in the Okareo app
        from_date (Union[Unset, datetime.datetime]): Earliest date Default: isoparse('2022-12-31T23:59:59.999999').
        to_date (Union[Unset, datetime.datetime]): Latest date
        feedback (Union[Unset, float]): Feedback is a 0 to 1 float value that captures user feedback range for related
            datapoint results
        error_code (Union[Unset, str]):
        context_token (Union[Unset, str]): Context token is a unique token to link various datapoints which originate
            from the same context
        project_id (Union[Unset, str]): Project ID
        mut_id (Union[Unset, str]): Model ID
        test_run_id (Union[Unset, str]): Test run ID
        search_value (Union[Unset, str]): Search substring that is matched against input, result, context_token, tags,
            or time_created fields
        offset (Union[Unset, int]): Offset for pagination
        limit (Union[Unset, int]): Limit for pagination
    """

    tags: Union[Unset, List[str]] = UNSET
    from_date: Union[Unset, datetime.datetime] = isoparse("2022-12-31T23:59:59.999999")
    to_date: Union[Unset, datetime.datetime] = UNSET
    feedback: Union[Unset, float] = UNSET
    error_code: Union[Unset, str] = UNSET
    context_token: Union[Unset, str] = UNSET
    project_id: Union[Unset, str] = UNSET
    mut_id: Union[Unset, str] = UNSET
    test_run_id: Union[Unset, str] = UNSET
    search_value: Union[Unset, str] = UNSET
    offset: Union[Unset, int] = UNSET
    limit: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        from_date: Union[Unset, str] = UNSET
        if not isinstance(self.from_date, Unset):
            from_date = self.from_date.isoformat()

        to_date: Union[Unset, str] = UNSET
        if not isinstance(self.to_date, Unset):
            to_date = self.to_date.isoformat()

        feedback = self.feedback
        error_code = self.error_code
        context_token = self.context_token
        project_id = self.project_id
        mut_id = self.mut_id
        test_run_id = self.test_run_id
        search_value = self.search_value
        offset = self.offset
        limit = self.limit

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if tags is not UNSET:
            field_dict["tags"] = tags
        if from_date is not UNSET:
            field_dict["from_date"] = from_date
        if to_date is not UNSET:
            field_dict["to_date"] = to_date
        if feedback is not UNSET:
            field_dict["feedback"] = feedback
        if error_code is not UNSET:
            field_dict["error_code"] = error_code
        if context_token is not UNSET:
            field_dict["context_token"] = context_token
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if test_run_id is not UNSET:
            field_dict["test_run_id"] = test_run_id
        if search_value is not UNSET:
            field_dict["search_value"] = search_value
        if offset is not UNSET:
            field_dict["offset"] = offset
        if limit is not UNSET:
            field_dict["limit"] = limit

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        tags = cast(List[str], d.pop("tags", UNSET))

        _from_date = d.pop("from_date", UNSET)
        from_date: Union[Unset, datetime.datetime]
        if isinstance(_from_date, Unset):
            from_date = UNSET
        else:
            from_date = isoparse(_from_date)

        _to_date = d.pop("to_date", UNSET)
        to_date: Union[Unset, datetime.datetime]
        if isinstance(_to_date, Unset):
            to_date = UNSET
        else:
            to_date = isoparse(_to_date)

        feedback = d.pop("feedback", UNSET)

        error_code = d.pop("error_code", UNSET)

        context_token = d.pop("context_token", UNSET)

        project_id = d.pop("project_id", UNSET)

        mut_id = d.pop("mut_id", UNSET)

        test_run_id = d.pop("test_run_id", UNSET)

        search_value = d.pop("search_value", UNSET)

        offset = d.pop("offset", UNSET)

        limit = d.pop("limit", UNSET)

        datapoint_search = cls(
            tags=tags,
            from_date=from_date,
            to_date=to_date,
            feedback=feedback,
            error_code=error_code,
            context_token=context_token,
            project_id=project_id,
            mut_id=mut_id,
            test_run_id=test_run_id,
            search_value=search_value,
            offset=offset,
            limit=limit,
        )

        datapoint_search.additional_properties = d
        return datapoint_search

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import time
from typing import Any, Optional
from unittest import mock
from unittest.mock import Mock

import pytest
from okareo_tests.common import random_string
from pytest_httpx import HTTPXMock

from okareo import Okareo


@pytest.fixture
def rnd() -> str:
    return random_string(5)


@pytest.fixture
def okareo_client() -> Okareo:
    return Okareo("foo", "http://mocked.com")


def get_mut_fixture(name: Optional[str] = None) -> dict:
    rnd_str = random_string(5)
    return {
        "id": "1",
        "project_id": "1",
        "name": name if name else f"CI-Async-Tests-{rnd_str}",
        "tags": ["ci-testing"],
        "time_created": "foo",
    }


def long_add_data_point(*k: Any, **kw: Any) -> None:
    time.sleep(1)


def test_non_blocking_call(okareo_client: Okareo, httpx_mock: HTTPXMock) -> None:
    fixture = get_mut_fixture()
    httpx_mock.add_response(status_code=201, json=fixture)

    mut = okareo_client.register_model(name=fixture["name"], tags=fixture["tags"])

    start = time.time()
    with mock.patch(
        "okareo.model_under_test.add_datapoint_v0_datapoints_post"
    ) as endpoint:
        endpoint.sync = Mock(side_effect=long_add_data_point)
        mut.add_data_point_async(
            feedback=0.1,
            context_token="SOME_CONTEXT_TOKEN",
        )
        # check if call was (almost) instant
        assert time.time() - start < 0.1
        # wait for long_add_data_point to execute
        time.sleep(2)
        assert endpoint.sync.called


def test_flush_leftovers(okareo_client: Okareo, httpx_mock: HTTPXMock) -> None:
    fixture = get_mut_fixture()
    httpx_mock.add_response(status_code=201, json=fixture)

    mut = okareo_client.register_model(name=fixture["name"], tags=fixture["tags"])

    start = time.time()
    with mock.patch(
        "okareo.model_under_test.add_datapoint_v0_datapoints_post"
    ) as endpoint:
        endpoint.sync = Mock(side_effect=long_add_data_point)
        mut.add_data_point_async(
            feedback=0.5,
            context_token="SOME_CONTEXT_TOKEN",
        )
        # should wait for long_add_data_point to comlpete
        mut.flush()
        assert time.time() - start > 1
        assert endpoint.sync.called


def test_retry_on_error(okareo_client: Okareo, httpx_mock: HTTPXMock) -> None:
    fixture = get_mut_fixture()
    httpx_mock.add_response(status_code=201, json=fixture)

    mut = okareo_client.register_model(name=fixture["name"], tags=fixture["tags"])

    time.time()
    with mock.patch(
        "okareo.model_under_test.add_datapoint_v0_datapoints_post"
    ) as endpoint:
        endpoint.sync = Mock(side_effect=Exception("Test"))
        mut.add_data_point_async(
            feedback=1,
            context_token="SOME_CONTEXT_TOKEN",
        )
        time.sleep(3)
        # there should be up to 5 retries
        assert endpoint.sync.call_count == 5


def test_send_once_above_queue_size_threshold(
    okareo_client: Okareo, httpx_mock: HTTPXMock
) -> None:
    fixture = get_mut_fixture()
    httpx_mock.add_response(status_code=201, json=fixture)

    mut = okareo_client.register_model(name=fixture["name"], tags=fixture["tags"])

    with mock.patch(
        "okareo.model_under_test.add_datapoint_v0_datapoints_post"
    ) as endpoint:
        with mock.patch("okareo.async_utils._DEFAULT_MAX_BATCH_SIZE", return_value=2):
            endpoint.sync = Mock()
            mut.add_data_point_async(
                feedback=0,
                context_token="SOME_CONTEXT_TOKEN",
            )
            assert not endpoint.sync.called
            mut.add_data_point_async(
                feedback=1,
                context_token="SOME_CONTEXT_TOKEN",
            )
            mut.add_data_point_async(
                feedback=2,
                context_token="SOME_CONTEXT_TOKEN",
            )

            assert not endpoint.sync.called



from pytest_httpx import HTTPXMock

from okareo import Okareo


def test_register_model(httpx_mock: HTTPXMock) -> None:
    fixture = {
        "id": "1",
        "project_id": "1",
        "name": "NotebookModel",
        "tags": ["ci-testing"],
        "time_created": "foo",
    }
    httpx_mock.add_response(status_code=201, json=fixture)
    okareo = Okareo("api-key", "http://mocked.com")
    mut = okareo.register_model(name="NotebookModel", tags=["ci-testing"])
    assert mut



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.datapoint_list_item_input_type_0 import DatapointListItemInputType0
    from ..models.datapoint_list_item_result_type_0 import DatapointListItemResultType0


T = TypeVar("T", bound="DatapointListItem")


@_attrs_define
class DatapointListItem:
    """
    Attributes:
        id (str):
        tags (Union[Unset, List[str]]):
        input_ (Union['DatapointListItemInputType0', List[Any], Unset, str]):
        input_datetime (Union[Unset, datetime.datetime]):
        result (Union['DatapointListItemResultType0', List[Any], Unset, str]):
        result_datetime (Union[Unset, datetime.datetime]):
        feedback (Union[Unset, float]):
        error_message (Union[Unset, str]):
        error_code (Union[Unset, str]):
        time_created (Union[Unset, datetime.datetime]):
        context_token (Union[Unset, str]):
        mut_id (Union[Unset, str]):
        project_id (Union[Unset, str]):
        test_run_id (Union[Unset, str]):
        total_search_count (Union[Unset, int]):
    """

    id: str
    tags: Union[Unset, List[str]] = UNSET
    input_: Union["DatapointListItemInputType0", List[Any], Unset, str] = UNSET
    input_datetime: Union[Unset, datetime.datetime] = UNSET
    result: Union["DatapointListItemResultType0", List[Any], Unset, str] = UNSET
    result_datetime: Union[Unset, datetime.datetime] = UNSET
    feedback: Union[Unset, float] = UNSET
    error_message: Union[Unset, str] = UNSET
    error_code: Union[Unset, str] = UNSET
    time_created: Union[Unset, datetime.datetime] = UNSET
    context_token: Union[Unset, str] = UNSET
    mut_id: Union[Unset, str] = UNSET
    project_id: Union[Unset, str] = UNSET
    test_run_id: Union[Unset, str] = UNSET
    total_search_count: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.datapoint_list_item_input_type_0 import DatapointListItemInputType0
        from ..models.datapoint_list_item_result_type_0 import DatapointListItemResultType0

        id = self.id
        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        input_: Union[Dict[str, Any], List[Any], Unset, str]
        if isinstance(self.input_, Unset):
            input_ = UNSET

        elif isinstance(self.input_, DatapointListItemInputType0):
            input_ = UNSET
            if not isinstance(self.input_, Unset):
                input_ = self.input_.to_dict()

        elif isinstance(self.input_, list):
            input_ = UNSET
            if not isinstance(self.input_, Unset):
                input_ = self.input_

        else:
            input_ = self.input_

        input_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.input_datetime, Unset):
            input_datetime = self.input_datetime.isoformat()

        result: Union[Dict[str, Any], List[Any], Unset, str]
        if isinstance(self.result, Unset):
            result = UNSET

        elif isinstance(self.result, DatapointListItemResultType0):
            result = UNSET
            if not isinstance(self.result, Unset):
                result = self.result.to_dict()

        elif isinstance(self.result, list):
            result = UNSET
            if not isinstance(self.result, Unset):
                result = self.result

        else:
            result = self.result

        result_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.result_datetime, Unset):
            result_datetime = self.result_datetime.isoformat()

        feedback = self.feedback
        error_message = self.error_message
        error_code = self.error_code
        time_created: Union[Unset, str] = UNSET
        if not isinstance(self.time_created, Unset):
            time_created = self.time_created.isoformat()

        context_token = self.context_token
        mut_id = self.mut_id
        project_id = self.project_id
        test_run_id = self.test_run_id
        total_search_count = self.total_search_count

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
            }
        )
        if tags is not UNSET:
            field_dict["tags"] = tags
        if input_ is not UNSET:
            field_dict["input"] = input_
        if input_datetime is not UNSET:
            field_dict["input_datetime"] = input_datetime
        if result is not UNSET:
            field_dict["result"] = result
        if result_datetime is not UNSET:
            field_dict["result_datetime"] = result_datetime
        if feedback is not UNSET:
            field_dict["feedback"] = feedback
        if error_message is not UNSET:
            field_dict["error_message"] = error_message
        if error_code is not UNSET:
            field_dict["error_code"] = error_code
        if time_created is not UNSET:
            field_dict["time_created"] = time_created
        if context_token is not UNSET:
            field_dict["context_token"] = context_token
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if test_run_id is not UNSET:
            field_dict["test_run_id"] = test_run_id
        if total_search_count is not UNSET:
            field_dict["total_search_count"] = total_search_count

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.datapoint_list_item_input_type_0 import DatapointListItemInputType0
        from ..models.datapoint_list_item_result_type_0 import DatapointListItemResultType0

        d = src_dict.copy()
        id = d.pop("id")

        tags = cast(List[str], d.pop("tags", UNSET))

        def _parse_input_(data: object) -> Union["DatapointListItemInputType0", List[Any], Unset, str]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _input_type_0 = data
                input_type_0: Union[Unset, DatapointListItemInputType0]
                if isinstance(_input_type_0, Unset):
                    input_type_0 = UNSET
                else:
                    input_type_0 = DatapointListItemInputType0.from_dict(_input_type_0)

                return input_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                input_type_1 = cast(List[Any], data)

                return input_type_1
            except:  # noqa: E722
                pass
            return cast(Union["DatapointListItemInputType0", List[Any], Unset, str], data)

        input_ = _parse_input_(d.pop("input", UNSET))

        _input_datetime = d.pop("input_datetime", UNSET)
        input_datetime: Union[Unset, datetime.datetime]
        if isinstance(_input_datetime, Unset):
            input_datetime = UNSET
        else:
            input_datetime = isoparse(_input_datetime)

        def _parse_result(data: object) -> Union["DatapointListItemResultType0", List[Any], Unset, str]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _result_type_0 = data
                result_type_0: Union[Unset, DatapointListItemResultType0]
                if isinstance(_result_type_0, Unset):
                    result_type_0 = UNSET
                else:
                    result_type_0 = DatapointListItemResultType0.from_dict(_result_type_0)

                return result_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                result_type_1 = cast(List[Any], data)

                return result_type_1
            except:  # noqa: E722
                pass
            return cast(Union["DatapointListItemResultType0", List[Any], Unset, str], data)

        result = _parse_result(d.pop("result", UNSET))

        _result_datetime = d.pop("result_datetime", UNSET)
        result_datetime: Union[Unset, datetime.datetime]
        if isinstance(_result_datetime, Unset):
            result_datetime = UNSET
        else:
            result_datetime = isoparse(_result_datetime)

        feedback = d.pop("feedback", UNSET)

        error_message = d.pop("error_message", UNSET)

        error_code = d.pop("error_code", UNSET)

        _time_created = d.pop("time_created", UNSET)
        time_created: Union[Unset, datetime.datetime]
        if isinstance(_time_created, Unset):
            time_created = UNSET
        else:
            time_created = isoparse(_time_created)

        context_token = d.pop("context_token", UNSET)

        mut_id = d.pop("mut_id", UNSET)

        project_id = d.pop("project_id", UNSET)

        test_run_id = d.pop("test_run_id", UNSET)

        total_search_count = d.pop("total_search_count", UNSET)

        datapoint_list_item = cls(
            id=id,
            tags=tags,
            input_=input_,
            input_datetime=input_datetime,
            result=result,
            result_datetime=result_datetime,
            feedback=feedback,
            error_message=error_message,
            error_code=error_code,
            time_created=time_created,
            context_token=context_token,
            mut_id=mut_id,
            project_id=project_id,
            test_run_id=test_run_id,
            total_search_count=total_search_count,
        )

        datapoint_list_item.additional_properties = d
        return datapoint_list_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.scenario_data_poin_response_input_type_0 import ScenarioDataPoinResponseInputType0
    from ..models.scenario_data_poin_response_result_type_0 import ScenarioDataPoinResponseResultType0


T = TypeVar("T", bound="ScenarioDataPoinResponse")


@_attrs_define
class ScenarioDataPoinResponse:
    """
    Attributes:
        id (str):
        input_ (Union['ScenarioDataPoinResponseInputType0', List[Any], str]):
        result (Union['ScenarioDataPoinResponseResultType0', List[Any], str]):
        meta_data (Union[Unset, str]):
    """

    id: str
    input_: Union["ScenarioDataPoinResponseInputType0", List[Any], str]
    result: Union["ScenarioDataPoinResponseResultType0", List[Any], str]
    meta_data: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.scenario_data_poin_response_input_type_0 import ScenarioDataPoinResponseInputType0
        from ..models.scenario_data_poin_response_result_type_0 import ScenarioDataPoinResponseResultType0

        id = self.id
        input_: Union[Dict[str, Any], List[Any], str]

        if isinstance(self.input_, ScenarioDataPoinResponseInputType0):
            input_ = self.input_.to_dict()

        elif isinstance(self.input_, list):
            input_ = self.input_

        else:
            input_ = self.input_

        result: Union[Dict[str, Any], List[Any], str]

        if isinstance(self.result, ScenarioDataPoinResponseResultType0):
            result = self.result.to_dict()

        elif isinstance(self.result, list):
            result = self.result

        else:
            result = self.result

        meta_data = self.meta_data

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "input": input_,
                "result": result,
            }
        )
        if meta_data is not UNSET:
            field_dict["meta_data"] = meta_data

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.scenario_data_poin_response_input_type_0 import ScenarioDataPoinResponseInputType0
        from ..models.scenario_data_poin_response_result_type_0 import ScenarioDataPoinResponseResultType0

        d = src_dict.copy()
        id = d.pop("id")

        def _parse_input_(data: object) -> Union["ScenarioDataPoinResponseInputType0", List[Any], str]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                input_type_0 = ScenarioDataPoinResponseInputType0.from_dict(data)

                return input_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                input_type_1 = cast(List[Any], data)

                return input_type_1
            except:  # noqa: E722
                pass
            return cast(Union["ScenarioDataPoinResponseInputType0", List[Any], str], data)

        input_ = _parse_input_(d.pop("input"))

        def _parse_result(data: object) -> Union["ScenarioDataPoinResponseResultType0", List[Any], str]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                result_type_0 = ScenarioDataPoinResponseResultType0.from_dict(data)

                return result_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                result_type_1 = cast(List[Any], data)

                return result_type_1
            except:  # noqa: E722
                pass
            return cast(Union["ScenarioDataPoinResponseResultType0", List[Any], str], data)

        result = _parse_result(d.pop("result"))

        meta_data = d.pop("meta_data", UNSET)

        scenario_data_poin_response = cls(
            id=id,
            input_=input_,
            result=result,
            meta_data=meta_data,
        )

        scenario_data_poin_response.additional_properties = d
        return scenario_data_poin_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
